#pragma once

#include "debug.hpp"

#include "dpt_types.hpp"

#include <ostream>
#include <utility>
#include <array>

/**
 * @file config.hpp Definition of all known dpt types.
 */

namespace knx {

template<int MAJOR>
class MappingMajor {};

#define KNX_MAP_MAJOR(major, size, type) \
template<> class MappingMajor<major> { public: enum { SIZE = size }; typedef type set_type; };

KNX_MAP_MAJOR(  1, 1,          bool );
KNX_MAP_MAJOR(  2, 1,   dpt_2::data );
KNX_MAP_MAJOR(  3, 1,   dpt_3::data );
KNX_MAP_MAJOR(  4, 2,          char );
KNX_MAP_MAJOR(  5, 2,       uint8_t );
KNX_MAP_MAJOR(  6, 2,        int8_t );
KNX_MAP_MAJOR(  7, 3,      uint16_t );
KNX_MAP_MAJOR(  8, 3,       int16_t );
KNX_MAP_MAJOR(  9, 3,         float );
KNX_MAP_MAJOR( 10, 4,  dpt_10::data );
KNX_MAP_MAJOR( 11, 4,  dpt_11::data );
KNX_MAP_MAJOR( 12, 5,      uint32_t );
KNX_MAP_MAJOR( 13, 5,       int32_t );
KNX_MAP_MAJOR( 14, 5,         float );
KNX_MAP_MAJOR( 15, 5,  dpt_15::data );
KNX_MAP_MAJOR( 16, 15, dpt_16::data );
KNX_MAP_MAJOR( 17, 2,  dpt_17::data );
KNX_MAP_MAJOR( 18, 2,  dpt_18::data );
KNX_MAP_MAJOR( 19, 8,  dpt_19::data );
KNX_MAP_MAJOR( 20, 2,       uint8_t );
KNX_MAP_MAJOR( 26, 2,  dpt_26::data );
KNX_MAP_MAJOR(219, 6, dpt_219::data );
KNX_MAP_MAJOR(232, 3, dpt_232::data );

//! \brief Forward declaration of hidden implementation.
template<int MAJOR, int MINOR>
class dpt_private;

//! \brief Template definition of each dpt.
template<int MAJOR, int MINOR>
class dpt
{
private:
    //! Pointer to hidden implementation details.
    dpt_private<MAJOR,MINOR> * dpt_private_pointer;
public:
    //! Fetching the set value type from mapping.
    typedef typename MappingMajor<MAJOR>::set_type set_type;

    //! Assigning the type for later access.
    typedef dpt<MAJOR,MINOR> own_type;

    //! Assigning type of private implementation.
    typedef dpt_private<MAJOR,MINOR> private_type;

    //! Constructor usually setting all values to zero.
    dpt();

    //! Destructor for the dpt.
    ~dpt();

    //! Function returning an index of the class for quick type access.
    static unsigned int get_index();

    //! Major function to set the value of a dpt.
    template<class TYPE>
    void set(const TYPE value);

    //! Major function to get the value of a dpt.
    template<class TYPE>
    void get(TYPE & value) const;

    //! Major function to output the value of a dpt.
    template<int MAJOR2, int MINOR2>
    friend std::ostream & operator<<(std::ostream & os, dpt<MAJOR2,MINOR2> & data);
};

//! Major function to output the value of a dpt. (see template of dpt)
template<int MAJOR, int MINOR>
std::ostream & operator<<(std::ostream & os, dpt<MAJOR,MINOR> & data);


//! Binary single bit value. Either on or off.
class dpt_1_001 : public dpt<1,001>
{
public:
    //! Static variable for switching on.
    KNX_IMPORT_EXPORT static const dpt<1,001>::set_type ON;
    //! Static variable for switching off.
    KNX_IMPORT_EXPORT static const dpt<1,001>::set_type OFF;
};

//! Binary single bit value with additional control bit. Either on or off.
class dpt_2_001 : public dpt<2,001>
{
public:
    //! Control bit is off so value is not used.
    KNX_IMPORT_EXPORT static const dpt<2,001>::set_type NO_CONTROL;
    //! Control bit is on and value is off.
    KNX_IMPORT_EXPORT static const dpt<2,001>::set_type CONTROL_VALUE_ZERO;
    //! Control bit is on and value is on.
    KNX_IMPORT_EXPORT static const dpt<2,001>::set_type CONTROL_VALUE_ONE;
};


//! Dimming control with direction and step value as 2^(value-1) subdivisions of the interval 0 to 100 percent.
class dpt_3_007 : public dpt<3,007>
{};


//! Simple single ascii character.
class dpt_4_001 : public dpt<4,001>
{};

//! Unsigned value with 8 bit. Scaling represents range 0 to 100 percent.

class dpt_5_001 : public dpt<5,001>
{};

//! Signed value with 8 bit. Percent represents range -128 to 127 percent.
class dpt_6_001 : public dpt<6,001>
{};

//! Unsigned value with 16 bit. Range from 0 to 65535.
class dpt_7_001 : public dpt<7,001>
{};

//! Signed value with 16 bit. Range from -32768 to 32767.
class dpt_8_001 : public dpt<8,001>
{};

//! Float value with 16 bit. Non IEEE definition. Range from -671088.64 to 670760.96. Here interpreted as temperature in degree celsius.
class dpt_9_001 : public dpt<9,001>
{
public:
    //! Invalid value.
    KNX_IMPORT_EXPORT static const uint16_t INVALID_DATA;
    //! Maximal value in binary.
    KNX_IMPORT_EXPORT static const uint16_t MAX;
    //! Minimal value in binary.
    KNX_IMPORT_EXPORT static const uint16_t MIN;
    //! Maximal value as float.
    KNX_IMPORT_EXPORT static const float MAX_FLOAT;
    //! Minimal value as float.
    KNX_IMPORT_EXPORT static const float MIN_FLOAT;
};

//! Representation of time with day, hours, minutes and seconds.
class dpt_10_001 : public dpt<10,001>
{
public:
    typedef dpt_10::data time;
};

//! Representation of date with day, month and year. Year has a value from 0 to 99 where values bigger than 90 are considered to be 1990+. Remark: Natural end in 2089!.
class dpt_11_001 : public dpt<11,001>
{
public:
    typedef dpt_11::data date;
};

//! Unsigned 32 bit. Here as counter.
class dpt_12_001 : public dpt<12,001> {};

//! Signed 32 bit. Here as counter.
class dpt_13_001 : public dpt<13,001> {};

//! Float as 32 bit with IEEE 754 notation. Here as acceleration in ms^-2.
class dpt_14_000 : public dpt<14,000> {};

//! Access data with six 4 byte fields.
class dpt_15_000 : public dpt<15,000> {};

//! An ascii string with max 14 ascii characters.
class dpt_16_000 : public dpt<16,000> {};

//! A 6 bit scene number.
class dpt_17_001 : public dpt<17,001> {};

//! A 6 bit scene number with control field.
class dpt_18_001 : public dpt<18,001> {};

//! Combinaton of date and time as 8 byte value.
class dpt_19_001 : public dpt<19,001> {};

//! Encoding of a status. Here autonomous(0), skave (1) and master (2).
class dpt_20_001 : public dpt<20,001> {};

//! Scene information with filed indicating if scene is active or not.
class dpt_26_001 : public dpt<26,001> {};

//! Alarm information.
class dpt_219_001 : public dpt<219,001> {};

//! Color information.
class dpt_232_600 : public dpt<232,600> {};

//! Alias for dpt 1.001
typedef dpt_1_001 dpt_switch;

//! Alias for dpt 10.001
typedef dpt_10_001 dpt_timeofday;

}
